--****************************************************************************
--**
--**  File     :  /lua/unit.lua
--**  Author(s):  John Comes, David Tomandl, Gordon Duclos
--**
--**  Summary  : The Unit lua module
--**
--**  Copyright  2005 Gas Powered Games, Inc.  All rights reserved.
--****************************************************************************

local oldUnit=Unit
Unit = Class(oldUnit) {

    OnCreate = function(self)
        oldUnit.OnCreate(self)
        local bp = self:GetBlueprint()
        if bp.Economy.XPperLevel then -- just to make certain this is not called
            --for units which do not need it
            self.XPnextLevel = bp.Economy.XPperLevel
            self.xp = 0
--            self.Txp = 0
            self.VeteranLevel = 1
            self.LevelProgress = 1
            -- removed BuildRate Sync Entry in order to speed up gameplay, there's an engine method we will use now.
--            self.Sync.XPnextLevel = self.XPnextLevel --Two new sync entries
--            self.Sync.Txp = self.Txp --hopefully will not slow things down
            self.Sync.LevelProgress = self.LevelProgress
            self.Sync.RegenRate = bp.Defense.RegenRate
        end
    end,

    OnStopBeingBuilt = function(self, builder, layer)
        if self:GetBlueprint().Economy.xpTimeStep then
            self:ForkThread(self.XPOverTime)
        end
        oldUnit.OnStopBeingBuilt(self, builder, layer)
    end,

--testcode
    --Economy-this is for everything which requires energy to be run: intel, shields etc
    SetMaintenanceConsumptionActive = function(self)
        if self.XPOverTimeThread then KillThread(self.XPOverTimeThread) end
        self.XPOverTimeThread = ForkThread(self.XPOverTime, self)
        oldUnit.SetMaintenanceConsumptionActive(self)
    end,

    SetMaintenanceConsumptionInactive = function(self)
        KillThread(self.XPOverTimeThread)
        oldUnit.SetMaintenanceConsumptionInactive(self)
    end,
--testcode end

    XPOverTime = function(self)
        if not self:GetBlueprint().Economy.xpTimeStep then return end -- sanity check
        local waittime = self:GetBlueprint().Economy.xpTimeStep * 0.5
        WaitSeconds(waittime) --wait longer to improve performance
        while not self:IsDead() do
            self:AddXP(self.XPnextLevel* 0.5)
            WaitSeconds(waittime)
        end
    end,

    startBuildXPThread = function(self)
        local levelPerSecond = self:GetBlueprint().Economy.BuildXPLevelpSecond
        if not levelPerSecond then return end
        WaitSeconds(1)--when you build THAT fast you do not need any xp, another performance tweak
        while not self:IsDead() do
            self:AddXP(self.XPnextLevel*levelPerSecond*0.01) --adjusted again, removed ceil, performance
            WaitSeconds(59) -- to reduce the performance impact
        end
    end,


--testcode


    -- reworked builder xp to include upgrades and allow for pausing
    -- launchers are still unaffected
    -- inconsistent gains for commanders.
    SetActiveConsumptionActive = function(self)
        if self.BuildXPThread then KillThread(self.BuildXPThread) end
        self.BuildXPThread = ForkThread(self.startBuildXPThread, self)
        oldUnit.SetActiveConsumptionActive(self)
    end,

    SetActiveConsumptionInactive = function(self)
        KillThread(self.BuildXPThread)
        if self.BuildXPThread then KillThread(self.BuildXPThread) end
        oldUnit.SetActiveConsumptionInactive(self)
    end,

    OnProductionPaused = function(self)
        KillThread(self.BuildXPThread)
        if self.BuildXPThread then KillThread(self.BuildXPThread) end
        oldUnit.OnProductionPaused(self)
    end,

    OnProductionUnpaused = function(self)
        if self.BuildXPThread then KillThread(self.BuildXPThread) end
        self.BuildXPThread = ForkThread(self.startBuildXPThread, self)
        oldUnit.OnProductionUnpaused(self)
    end,

--end testcode


--    OnStartBuild = function(self, unitBeingBuilt, order)
--        --LOG('unit:OnStartBuild')
--        self.BuildXPThread = ForkThread(self.startBuildXPThread, self)
--        oldUnit.OnStartBuild(self, unitBeingBuilt, order)
--    end,
--
    OnStopBuild = function(self, unitBeingBuilt)
        --LOG('unit:OnStopBuild')
        KillThread(self.BuildXPThread)
        oldUnit.OnStopBuild(self, unitBeingBuilt)
    end,

    OnFailedToBuild = function(self)
        --LOG('unit:OnFailedToBuild')
        KillThread(self.BuildXPThread)
        oldUnit.OnFailedToBuild(self)
    end,


    ------------------------------------------------------------------------------------- VETERANCY
    ---------------------------------------------------------------------------------

    -- use this to go through the AddKills function rather than directly setting veterancy
    SetVeterancy = function(self, veteranLevel)
        veteranLevel = veteranLevel or 0
        if veteranLevel <= 5 then
            return oldUnit.SetVeterancy(self, veteranLevel)
        else
            local bp = self:GetBlueprint()-- change suggested by Gilbot
            if bp.Veteran['Level'..veteranLevel] then
                self:AddKills(bp.Veteran['Level'..veteranLevel])
            else
                WARN('SetVeterancy called on ' .. self:GetUnitId()
                .. ' with veteran level ' .. veteranLevel
                .. ' which was not defined in its BP file. '
                .. ' Veterancy level has not been set.'
                )
            end
        end
    end,

    --Check to see if we should veteran up.
    CheckVeteranLevel = function(self)
        if not self.XPnextLevel then return end
        local bp = self:GetBlueprint()
--        LOG('xp:' .. self.xp.. ' level at:' .. self.XPperLevel)
        while self.xp >= self.XPnextLevel do
            self.xp = self.xp - self.XPnextLevel
            self.XPnextLevel = bp.Economy.XPperLevel * (1+ 0.1*self.VeteranLevel)
            self:SetVeteranLevel(self.VeteranLevel + 1)--fix this, it is causing  overhead
        end
        self.LevelProgress = self.xp / self.XPnextLevel + self.VeteranLevel
        self.Sync.LevelProgress = self.LevelProgress
--        self.Sync.XPnextLevel = self.Txp - self.xp + self.XPnextLevel
        -- syncing XP need for Levelup, calculation within the UI is faulty.
    end,

    --fix, 117 did not adjust this to reflect new balance
    AddLevels = function(self, levels)
        local bp = self:GetBlueprint()
        local curlevel = self.VeteranLevel
        local percent = self.LevelProgress - curlevel
        local xpAdd = 0
        if levels >= (1-percent) then
            xpAdd = self.XPnextLevel * (1-percent)
            levels=levels-(1-percent)
        else
            xpAdd =self.XPnextLevel * levels
            levels=0
        end
        while levels > 1 do
            levels=levels-1
            curlevel = curlevel +1
            xpAdd=xpAdd + bp.Economy.XPperLevel * (1+ 0.1*curlevel) --fixed
        end -- to account for the 10% increase per gained level
        xpAdd=xpAdd + bp.Economy.XPperLevel * (1+ 0.1*(curlevel+1)) * levels
        self:AddXP(xpAdd) --same
    end,

    --Set the veteran level to the level specified
    SetVeteranLevel = function(self, level)
        --LOG(' ')
        --LOG('*DEBUG: '.. self:GetBlueprint().Description .. ' VETERAN UP! LEVEL ', repr(level))
        local old = self.VeteranLevel
        self.VeteranLevel = level

        -- Apply default veterancy buffs
        local buffTypes = { 'Regen', 'Health', 'Damage','DamageArea','Range','Speed','Vision','OmniRadius','Radar','BuildRate','ResourceProduction','RateOfFire','Shield'}
        for k,bType in buffTypes do
            Buff.ApplyBuff( self, 'Veterancy' .. bType)
            --Buff.ApplyBuff( self, 'Veterancy' .. bType .. level )
        end

        -- Get any overriding buffs if they exist
        local bp = self:GetBlueprint().Buffs
        --Check for unit buffs

        if bp then
            for bLevel,bData in bp do
                if (bLevel == 'Any' or bLevel == 'Level'..level) then
                    for bType,bValues in bData do
                        local buffName = self:CreateUnitBuff(bLevel,bType,bValues)
                        if buffName then
                            Buff.ApplyBuff( self, buffName )
                        end
                    end
                end
            end
        end

        self:GetAIBrain():OnBrainUnitVeterancyLevel(self, level)
        self:DoUnitCallbacks('OnVeteran')
    end,

    --function to generate the new veterancy buffs
    CreateUnitBuff = function(self, levelName, buffType, buffValues)

        -- Generate a buff based on the unitId
        local buffName = self:GetUnitId() .. levelName .. buffType
        local buffMinLevel = nil
        local buffMaxLevel = nil
        if buffValues.MinLevel then buffMinLevel = buffValues.MinLevel end
        if buffValues.MaxLevel then buffMaxLevel = buffValues.MaxLevel end


        -- Create the buff if needed
        if not Buffs[buffName] then
            --LOG(buffName .. ': '..buffMinLevel.. ' - '..buffMaxLevel)
            BuffBlueprint {
                MinLevel = buffMinLevel,
                MaxLevel = buffMaxLevel,
                Name = buffName,
                DisplayName = buffName,
                BuffType = buffType,
                Stacks = buffValues.Stacks,
                --self.BuffTypes[buffType].BuffStacks,
                Duration = buffValues.Duration,
                Affects = buffValues.Affects,
            }
        end

        -- Return the buffname so the buff can be applied to the unit
        return buffName
    end,

    --Allowing buff bonus and adj bonus at same time!!
    UpdateProductionValues = function(self)
        local bpEcon = self:GetBlueprint().Economy
        if not bpEcon then return end
        self:SetProductionPerSecondEnergy( (self.EnergyProdMod or bpEcon.ProductionPerSecondEnergy or 0)* (self.EnergyProdAdjMod or 1) )
        self:SetProductionPerSecondMass( (self.MassProdMod or bpEcon.ProductionPerSecondMass or 0) * (self.MassProdAdjMod or 1) )
    end,

    AddXP = function(self,amount)
        if not self.XPnextLevel then return end
--        self.Txp = self.Txp + (math.ceil(amount))
--        self.Sync.Txp = self.Txp-- Total EXP collected
        self.xp = self.xp + (amount)--removed ceil, hope this speeds things up
--        LOG('___' .. amount .. repr(self:GetBlueprint().Description))
        self:CheckVeteranLevel()
    end,

    DoTakeDamage = function(self, instigator, amount, vector, damageType)
        if instigator and IsUnit(instigator) and not instigator:IsDead() then
            local bp = self:GetBlueprint()
            if bp.Economy.xpPerHp then
                local preAdjHealth = self:GetHealth()
                if amount>=preAdjHealth then
                    local sLevel = self.VeteranLevel
--                    LOG('___sL: ' .. sLevel)
                    local xpR = sLevel
                    instigator:AddXP(bp.Economy.xpValue*xpR*0.25)
                    --xp gains only on killing a unit
--                    LOG('___xpR: ' .. xpR  )
--                    LOG('___xpV: ' .. bp.Economy.xpValue  )
--                    LOG('___xpT: ' .. (bp.Economy.xpValue*xpR*AoE))
                end
            end
        end
        oldUnit.DoTakeDamage(self, instigator, amount, vector, damageType)
    end,

    GetShield = function(self)
        return self.MyShield or nil
    end,
}

